# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
Helper for relation/customer resources.

This file is part of the Exact Online REST API Library in Python
(EORALP), licensed under the LGPLv3+.
Copyright (C) 2015 Walter Doekes, OSSO B.V.
"""
from .manager import Manager


class Relations(Manager):
    resource = 'crm/Accounts'

    def filter(self, relation_code=None, **kwargs):
        # $select=ID,Code,Name
        if 'select' not in kwargs:
            kwargs['select'] = 'ID,Code,Name'

        if relation_code is not None:
            remote_id = self._remote_relation_code(relation_code)
            self._filter_append(kwargs, u'Code eq %s' % (remote_id,))
        return super(Relations, self).filter(**kwargs)

    def _remote_relation_code(self, code):
        return u"'%18s'" % (code.replace("'", "''"),)



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
Provides a storage class to the Exact Online REST API Library.

This file is part of the Exact Online REST API Library in Python
(EORALP), licensed under the LGPLv3+.
Copyright (C) 2015-2016 Walter Doekes, OSSO B.V.

Usage:

    storage = IniStorage('somefile.ini')

Or, if you want to use your own storage backend, you may provide a set()
and get() method and inherit from the ExactOnlineConfig.

    class MyStorage(ExactOnlineConfig):
        def get(self, section, option):
            try:
                return self.fetch_stuff_from_somewhere(...)
            except:
                raise NoOptionError()

        def set(self, section, option, value):
            self.store_stuff_somewhere(...)

Example ini file:

    [server]
    auth_url = https://start.exactonline.co.uk/api/oauth2/auth
    rest_url = https://start.exactonline.co.uk/api
    token_url = https://start.exactonline.co.uk/api/oauth2/token

    [application]
    base_url = https://example.com
    client_id = {12345678-abcd-1234-abcd-0123456789ab}
    client_secret = ZZZ999xxx000

    [transient]
    access_expiry = 1426492503
    access_token = dAfjGhB1k2tE2dkG12sd1Ff1A1fj2fH2Y1j1fKJl2f1sD1ON275zJNUy...
    code = dAfj!hB1k2tE2dkG12sd1Ff1A1fj2fH2Y1j1fKJl2f1sD1ON275zJNUy...
    division = 123456
    refresh_token = SDFu!12SAah-un-56su-1fj2fH2Y1j1fKJl2f1sDfKJl2f1sD11FfUn1...

"""
try:
    from configparser import NoSectionError, NoOptionError, ConfigParser
except ImportError:  # python2
    from ConfigParser import (
        NoSectionError, NoOptionError,
        SafeConfigParser as ConfigParserOldStyle)

    class ConfigParser(ConfigParserOldStyle, object):
        """
        We require this adapter to upgrade the RawConfigParser to a
        new-style class. Only needed in Python2.
        """
        def __init__(self, **kwargs):
            # Must call this __init__ manually :(
            ConfigParserOldStyle.__init__(self, **kwargs)
            super(ConfigParser, self).__init__(**kwargs)

try:
    unicode
except NameError:
    # python3
    def native_string(value):
        if isinstance(value, str):
            return value
        if isinstance(value, bytes):
            return value.decode('utf-8')
        return str(value)
else:
    # python2
    def native_string(value):
        if isinstance(value, str):
            return value
        if isinstance(value, unicode):
            return value.encode('utf-8')
        return str(value)


# TODO: replace NoOptionError with a custom/own config error?
# TODO: move the exactonlineconfig base to a separate file one doesn't need to
# import from the ConfigParser?


class ExactOnlineConfig(object):
    """
    The configuration class.

    The getters all assert that a native-string type is returned (unless
    a different type is explicitly mentioned). This way the callers
    won't have to think about what they get.
    """
    def get_or_set_default(self, section, option, value):
        """
        Base method to fetch values and to set defaults in case they
        don't exist.
        """
        try:
            ret = self.get(section, option)
        except (NoOptionError, NoSectionError):
            self.set(section, option, value)
            ret = value

        return ret

    # [server]
    # ; These always return something. They may write their default to
    # ; your config object if they are unset.
    # ; If you want to change the values from your application, you need
    # ; to write your own routines.

    def get_auth_url(self):
        return native_string(self.get_or_set_default(
            'server', 'auth_url',
            'https://start.exactonline.nl/api/oauth2/auth'))

    def get_rest_url(self):
        return native_string(self.get_or_set_default(
            'server', 'rest_url',
            'https://start.exactonline.nl/api'))

    def get_token_url(self):
        return native_string(self.get_or_set_default(
            'server', 'token_url',
            'https://start.exactonline.nl/api/oauth2/token'))

    # [application]
    # ; These return something, or raise NoOptionError if they are
    # ; unset.
    # ; If you want to change the values from your application, you need
    # ; to write your own routines.

    def get_base_url(self):
        return native_string(self.get('application', 'base_url'))

    def get_client_id(self):
        return native_string(self.get('application', 'client_id'))

    def get_client_secret(self):
        return native_string(self.get('application', 'client_secret'))

    # [transient]
    # ; These return something, or raise NoOptionError if they are
    # ; unset.
    # ; These will be written to by this exactonline library, so all
    # ; their get methods have set counterparts.

    def get_access_expiry(self):
        return int(self.get('transient', 'access_expiry'))

    def set_access_expiry(self, value):
        self.set('transient', 'access_expiry', native_string(value))

    def get_access_token(self):
        return native_string(self.get('transient', 'access_token'))

    def set_access_token(self, value):
        self.set('transient', 'access_token', native_string(value))

    def get_code(self):
        return native_string(self.get('transient', 'code'))

    def set_code(self, value):
        self.set('transient', 'code', native_string(value))

    def get_division(self):
        return int(self.get('transient', 'division'))

    def set_division(self, value):
        self.set('transient', 'division', native_string(value))

    def get_refresh_token(self):
        return native_string(self.get('transient', 'refresh_token'))

    def set_refresh_token(self, value):
        self.set('transient', 'refresh_token', native_string(value))

    # ; aliases

    def get_refresh_url(self):
        "Alias for get_token_url()."
        return self.get_token_url()


class IniStorage(ExactOnlineConfig, ConfigParser):
    """
    Configuration based on the SafeConfigParser and the
    ExactOnlineConfig.

    Takes a ``filename_or_fp`` which can either be a filename or a file
    pointer. If it is a filename, all set() operations are destructive:
    the file will be automatically updated.
    """
    def __init__(self, filename_or_fp, **kwargs):
        super(IniStorage, self).__init__(**kwargs)

        if hasattr(filename_or_fp, 'read'):
            if hasattr(self, 'read_file'):
                self.read_file(filename_or_fp)
            else:
                self.readfp(filename_or_fp)  # python<3.2
            self.overwrite = False
        else:
            self.read(filename_or_fp)
            self.overwrite = filename_or_fp

    def get(self, section, option, **kwargs):
        """
        Get method that raises NoOptionError if the value was unset.
        This differs from the SafeConfigParser which may also raise a
        NoSectionError.

        We take extra **kwargs because the Python 3.5 configparser extends the
        get method signature and it calls self with those parameters.

            def get(self, section, option, *, raw=False, vars=None,
                    fallback=_UNSET):
        """
        try:
            ret = super(ExactOnlineConfig, self).get(section, option, **kwargs)
        except NoSectionError:
            raise NoOptionError(option, section)

        return ret

    def set(self, section, option, value):
        """
        Set method that (1) auto-saves if possible and (2) auto-creates
        sections.
        """
        try:
            super(ExactOnlineConfig, self).set(section, option, value)
        except NoSectionError:
            self.add_section(section)
            super(ExactOnlineConfig, self).set(section, option, value)

        # Save automatically!
        self.save()

    def save(self):
        if self.overwrite:
            with open(self.overwrite, 'w') as output:
                self.write(output)



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
Data structures for communication with remote.

Example usage::

    class BossoExactCustomer(ExactCustomer):
        def __init__(self, bosso_relation=None, **kwargs):
            super(BossoExactRelation, self).__init__(**kwargs)

            self._bosso_relation = bosso_relation

        def get_code(self):
            return self._bosso_relation.code

        def get_name(self):
            return self._bosso_relation.name

        def get_address(self):
            address = self._bosso_relation.billing_address
            if address:
                return {
                    'AddressLine1': address.street_and_number(),
                    'Postcode': address.zipcode,
                    'City': address.city.name,
                }
            return {}

This file is part of the Exact Online REST API Library in Python
(EORALP), licensed under the LGPLv3+.
Copyright (C) 2015 Walter Doekes, OSSO B.V.
"""
from .base import ExactElement
from ..exceptions import ObjectDoesNotExist


class ExactRelation(ExactElement):
    def get_guid(self):
        exact_relation = self._api.relations.get(relation_code=self.get_code())
        return exact_relation['ID']

    def get_code(self):
        # The "foreign"/local relation code.
        raise NotImplementedError()

    def get_name(self):
        raise NotImplementedError()

    def get_address(self):
        # Return a dict, like:
        # {'AddressLine1': 'x', 'Postcode': 'y',
        #  'City': 'z'}
        return {}

    def assemble(self):
        # Compile a dict to be pushed.
        raise NotImplementedError()

    def commit(self):
        try:
            exact_guid = self.get_guid()
        except ObjectDoesNotExist:
            exact_guid = None

        data = self.assemble()

        if exact_guid:
            ret = self._api.relations.update(exact_guid, data)
            # ret is None
        else:
            ret = self._api.relations.create(data)
            # ret is a exact_relation

        return ret


class ExactCustomer(ExactRelation):
    def assemble(self):
        data = {
            # 18 spaces for code.
            'Code': '%18s' % (self.get_code(),),
            'Name': self.get_name(),
            # This speaks for itself.
            'PaymentConditionSales': 15,
            'PaymentConditionSalesDescription': 'Binnen 15 dagen',
            # This defaults to 1, no idea what this is.
            'InvoiceAttachmentType': 0,
            # Status is: A=supplier, C=customer (I think)
            # (suppliers have IsPurchase and IsSales as False and have
            # IsSupplier as True instead)
            'Status': 'C',
            'IsPurchase': True,
            'IsSales': True,
        }

        # Billing address.
        data.update(self.get_address())

        # Example data:
        # {'AccountManager': None,
        #  'AccountManagerFullName': None,
        #  'AccountManagerHID': None,
        #  'Accountant': None,
        #  'ActivitySector': None,
        #  'ActivitySubSector': None,
        #  'AddressLine1': 'SomeAddress 1',
        #  'AddressLine2': None,
        #  'AddressLine3': None,
        #  'BRIN': None,
        #  'BankAccounts': {u'__deferred':
        #    {u'uri': "https://start.exactonline.nl/api/v1/543210/\
        #              crm/Accounts(guid'eb7...600')/BankAccounts"}},
        #  'Blocked': False,
        #  'BusinessType': None,
        #  'CanDropShip': False,
        #  'ChamberOfCommerce': None,
        #  'City': 'SomeCity',
        #  'Classification': None,
        #  'ClassificationDescription': None,
        #  'Code': '               255',
        #  'CodeAtSupplier': None,
        #  'CompanySize': None,
        #  'ConsolidationScenario': 2,
        #  'ControlledDate': None,
        #  'CostPaid': 0,
        #  'Costcenter': None,
        #  'CostcenterDescription': None,
        #  'Country': 'NL ',
        #  'Created': '/Date(1392224884577)/',
        #  'Creator': 'e17...1cb',
        #  'CreatorFullName': 'Her...Bos',
        #  'CreditLinePurchase': 0,
        #  'CreditLineSales': 0,
        #  'Currency': None,
        #  'CustomerSince': None,
        #  'DiscountPurchase': 0,
        #  'DiscountSales': 0,
        #  'Division': 543210,
        #  'Document': None,
        #  'DunsNumber': None,
        #  'Email': None,
        #  'EndDate': None,
        #  'Fax': None,
        #  'GLAP': None,
        #  'GLAR': None,
        #  'GLAccountPurchase': None,
        #  'GLAccountSales': None,
        #  'ID': 'eb7...600',
        #  'IntraStatArea': None,
        #  'IntraStatDeliveryTerm': None,
        #  'IntraStatSystem': None,
        #  'IntraStatTransactionA': None,
        #  'IntraStatTransactionB': None,
        #  'IntraStatTransportMethod': None,
        #  'InvoiceAccount': None,
        #  'InvoiceAccountCode': None,
        #  'InvoiceAccountName': None,
        #  'InvoiceAttachmentType': 0,
        #  'InvoicingMethod': 0,
        #  'IsAccountant': 0,
        #  'IsAgency': 0,
        #  'IsBank': False,
        #  'IsCompetitor': 0,
        #  'IsMailing': 0,
        #  'IsMember': False,
        #  'IsPilot': False,
        #  'IsPurchase': True,
        #  'IsReseller': False,
        #  'IsSales': True,
        #  'IsSupplier': False,
        #  'Language': None,
        #  'LanguageDescription': None,
        #  'Latitude': None,
        #  'LeadSource': None,
        #  'Logo': None,
        #  'LogoFileName': None,
        #  'LogoThumbnailUrl': 'https://start.exactonline.nl//docs/images/\
        #                       placeholder_account_myeol.png',
        #  'LogoUrl': 'https://start.exactonline.nl//docs/images/\
        #              placeholder_account_myeol.png',
        #  'Longitude': None,
        #  'MainContact': None,
        #  'Modified': '/Date(1392228726777)/',
        #  'Modifier': 'e17...1cb',
        #  'ModifierFullName': 'Her...Bos',
        #  'Name': 'SOME CUSTOMER',
        #  'Parent': None,
        #  'PaymentConditionPurchase': None,
        #  'PaymentConditionPurchaseDescription': None,
        #  'PaymentConditionSales': '15',
        #  'PaymentConditionSalesDescription': 'Binnen 15 dagen',
        #  'Phone': None,
        #  'PhoneExtension': None,
        #  'Postcode': '1111 AA',
        #  'PriceList': None,
        #  'PurchaseCurrency': 'EUR',
        #  'PurchaseCurrencyDescription': 'Euro',
        #  'PurchaseLeadDays': 0,
        #  'PurchaseVATCode': None,
        #  'PurchaseVATCodeDescription': None,
        #  'RecepientOfCommissions': False,
        #  'Remarks': None,
        #  'Reseller': None,
        #  'ResellerCode': None,
        #  'ResellerName': None,
        #  'SalesCurrency': 'EUR',
        #  'SalesCurrencyDescription': 'Euro',
        #  'SalesVATCode': None,
        #  'SalesVATCodeDescription': None,
        #  'SearchCode': None,
        #  'SecurityLevel': 10,
        #  'SeparateInvPerProject': 0,
        #  'SeparateInvPerSubscription': 0,
        #  'ShippingLeadDays': 0,
        #  'ShippingMethod': None,
        #  'StartDate': '/Date(1392163200000)/',
        #  'State': None,
        #  'StateName': None,
        #  'Status': 'C',
        #  'StatusSince': '/Date(1392163200000)/',
        #  'Type': 'A',
        #  'VATLiability': None,
        #  'VATNumber': None,
        #  'Website': None,
        #  '__metadata': {u'type': 'Exact.Web.Api.Models.Account',
        #                  'uri': "https://start.exactonline.nl/api/v1/543210/\
        #                          crm/Accounts(guid'eb7...600')"}}

        return data



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
Combines the helper superclasses and the helper resource managers into the
ExactApi class.

This file is part of the Exact Online REST API Library in Python
(EORALP), licensed under the LGPLv3+.
Copyright (C) 2015 Walter Doekes, OSSO B.V.
"""
from ..rawapi import ExactRawApi

from .autorefresh import Autorefresh
from .unwrap import Unwrap
from .v1division import V1Division

from .invoices import Invoices
from .ledgeraccounts import LedgerAccounts
from .receivables import Receivables
from .relations import Relations


class ExactApi(
    # Talk to /api/v1/{division} directly.
    V1Division,
    # Strip the surrounding "d" and "results" dictionary
    # items.
    Unwrap,
    # Ensure that tokens are refreshed: if we get a 401, refresh the
    # tokens.
    Autorefresh,
    # The base class comes last: talk to /api.
    ExactRawApi
):
    invoices = Invoices.as_property()
    ledgeraccounts = LedgerAccounts.as_property()
    receivables = Receivables.as_property()
    relations = Relations.as_property()



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
Helper for invoice resources.

This file is part of the Exact Online REST API Library in Python
(EORALP), licensed under the LGPLv3+.
Copyright (C) 2015-2016 Walter Doekes, OSSO B.V.
"""
from .manager import Manager


class Invoices(Manager):
    resource = 'salesentry/SalesEntries'

    def get(self, **kwargs):
        invoice_dict = super(Invoices, self).get(**kwargs)
        try:
            uri = invoice_dict[u'SalesEntryLines'][u'__deferred']['uri']
        except KeyError:
            # Perhaps there is a 'select' filter.
            pass
        else:
            invoicelines_dict = self._api.restv1('GET', str(uri))
            invoice_dict[u'SalesEntryLines'] = invoicelines_dict
        return invoice_dict

    def filter(self, invoice_number=None, invoice_number__in=None,
               reporting_period=None, **kwargs):
        if invoice_number is not None:
            remote_id = self._remote_invoice_number(invoice_number)
            # Filter by our invoice_number.
            self._filter_append(kwargs, u'YourRef eq %s' % (remote_id,))
            # # Let the query return the invoice lines too. <-- DOES NOT WORK
            # assert 'expand' not in kwargs
            # kwargs['expand'] = 'SalesInvoiceLines'
            # kwargs['select'] = ('Amount,SalesInvoiceLines/LineNumber,'
            #                     'SalesInvoiceLines/Amount,'
            #                     'SalesInvoiceLines/VATAmount')

        if invoice_number__in is not None:
            # Filter by any of the supplied invoice numbers.
            remote_filter = []
            for invoice_number in invoice_number__in:
                remote_id = self._remote_invoice_number(invoice_number)
                remote_filter.append(u'YourRef eq %s' % (remote_id,))
            self._filter_append(kwargs, u'(%s)' % (u' or '.join(remote_filter),))

        if reporting_period is not None:
            # Filter by reporting period.
            self._filter_append(
                kwargs, u'ReportingYear eq %d' % (reporting_period.year,))
            self._filter_append(
                kwargs, u'ReportingPeriod eq %d' % (reporting_period.month,))

        return super(Invoices, self).filter(**kwargs)

    def map_exact2foreign_invoice_numbers(self, exact_invoice_numbers=None):
        """
        Optionally supply a list of ExactOnline invoice numbers.

        Returns a dictionary of ExactOnline invoice numbers to foreign
        (YourRef) invoice numbers.
        """
        # Quick, select all. Not the most nice to the server though.
        if exact_invoice_numbers is None:
            ret = self.filter(select='InvoiceNumber,YourRef')
            return dict((i['InvoiceNumber'], i['YourRef']) for i in ret)

        # Slower, select what we want to know. More work for us.
        exact_to_foreign_map = {}

        # Do it in batches. If we append 300 InvoiceNumbers at once, we
        # get a 12kB URI. (If the list is empty, we skip the entire
        # forloop and correctly return the empty dict.)
        exact_invoice_numbers = list(set(exact_invoice_numbers))  # unique
        for offset in range(0, len(exact_invoice_numbers), 40):
            batch = exact_invoice_numbers[offset:(offset + 40)]
            filter_ = ' or '.join(
                'InvoiceNumber eq %s' % (i,) for i in batch)
            assert filter_  # if filter was empty, we'd get all!
            ret = self.filter(filter=filter_, select='InvoiceNumber,YourRef')
            exact_to_foreign_map.update(
                dict((i['InvoiceNumber'], i['YourRef']) for i in ret))

        # Any values we missed?
        for exact_invoice_number in exact_invoice_numbers:
            if exact_invoice_number not in exact_to_foreign_map:
                exact_to_foreign_map[exact_invoice_number] = None

        return exact_to_foreign_map

    def map_foreign2exact_invoice_numbers(self, foreign_invoice_numbers=None):
        """
        Optionally supply a list of foreign (your) invoice numbers.

        Returns a dictionary of your invoice numbers (YourRef) to Exact
        Online invoice numbers.
        """
        # Quick, select all. Not the most nice to the server though.
        if foreign_invoice_numbers is None:
            ret = self.filter(select='InvoiceNumber,YourRef')
            return dict((i['YourRef'], i['InvoiceNumber']) for i in ret)

        # Slower, select what we want to know. More work for us.
        foreign_to_exact_map = {}

        # Do it in batches. If we append 300 InvoiceNumbers at once, we
        # get a 12kB URI. (If the list is empty, we skip the entire
        # forloop and correctly return the empty dict.)
        foreign_invoice_numbers = list(set(foreign_invoice_numbers))  # unique
        for offset in range(0, len(foreign_invoice_numbers), 40):
            batch = foreign_invoice_numbers[offset:(offset + 40)]
            filter_ = ' or '.join(
                'YourRef eq %s' % (self._remote_invoice_number(i),)
                for i in batch)
            assert filter_  # if filter was empty, we'd get all!
            ret = self.filter(filter=filter_, select='InvoiceNumber,YourRef')
            foreign_to_exact_map.update(
                dict((i['YourRef'], i['InvoiceNumber']) for i in ret))

        # Any values we missed?
        for foreign_invoice_number in foreign_invoice_numbers:
            if foreign_invoice_number not in foreign_to_exact_map:
                foreign_to_exact_map[foreign_invoice_number] = None

        return foreign_to_exact_map

    def _remote_invoice_number(self, invoice_number):
        return u"'%s'" % (invoice_number.replace("'", "''"),)



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
Base manager class for resource helpers.

This file is part of the Exact Online REST API Library in Python
(EORALP), licensed under the LGPLv3+.
Copyright (C) 2015 Walter Doekes, OSSO B.V.
"""
from ..exceptions import MultipleObjectsReturned, ObjectDoesNotExist
from ..http import binquote


# Python23 compatibility helpers
try:
    unicode  # python2 only
except NameError:
    to_binstr = (lambda x: x.encode('utf-8'))  # unistr-to-binstr
    to_unistr = str  # nonstr-to-unistr
else:
    to_binstr = str
    to_unistr = unicode  # noqa: non-str-to-unistr


class Manager(object):
    """
    Inherit from this when you're creating a property on the ExactApi,
    like this:

        class Relations(Manager):
            resource = 'some/resource'

        class ExactApi(...):
            relations = Relations.as_property()

        storage = SomeStorage()
        api = ExactApi(storage=storage)
        api.relations.all()
    """
    resource = None  # set this in your subclass

    @classmethod
    def as_property(cls):
        @property
        def cached_getter(exactapi):
            propname = '_prop_%s' % (cls.__name__,)
            if not hasattr(exactapi, propname):
                setattr(exactapi, propname, cls(exactapi))
            return getattr(exactapi, propname)

        return cached_getter

    def __init__(self, api):
        self._api = api

    # == GET / get one / get many ==

    def all(self):
        # Select all without filtering.
        return self.filter()

    def get(self, **kwargs):
        assert 'top' not in kwargs
        ret = self.filter(top=2, **kwargs)
        if not ret:
            raise ObjectDoesNotExist()
        if len(ret) > 1:
            raise MultipleObjectsReturned()
        return ret[0]

    def filter(self, **kwargs):
        # kwargs = {'filter': "EntryDate+gt+datetime'2014-01-01'", 'top': 5}
        args = []
        for key, value in kwargs.items():
            args.append('$%s=%s' % (
                key, binquote(to_unistr(value))))
        if args:
            args = ('?' + '&'.join(args))
        else:
            args = ''

        ret = self._api.restv1('GET', self.resource + args)
        return ret

    # == POST / create ==

    def create(self, element_dict):
        ret = self._api.restv1('POST', str(self.resource), element_dict)
        return ret

    # == DELETE / remove ==

    def delete(self, remote_guid):
        remote_id = self._remote_guid(remote_guid)
        uri = '%s(%s)' % (self.resource, remote_id)
        ret = self._api.restv1('DELETE', str(uri))
        return ret

    # == PUT / update ==

    def update(self, remote_guid, element_dict):
        remote_id = self._remote_guid(remote_guid)
        uri = '%s(%s)' % (self.resource, remote_id)
        ret = self._api.restv1('PUT', str(uri), element_dict)
        return ret

    # == helpers ==

    def _filter_append(self, kwargs, extra_filter):
        if 'filter' in kwargs:
            kwargs['filter'] = u'(%s) and %s' % (kwargs['filter'],
                                                 extra_filter)
        else:
            kwargs['filter'] = extra_filter

    def _remote_datetime(self, remote_datetime):
        return "datetime'%s'" % (remote_datetime.strftime('%Y-%m-%d'),)

    def _remote_guid(self, remote_guid):
        return "guid'%s'" % (remote_guid,)



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
Data structures for communication with remote.

This file is part of the Exact Online REST API Library in Python
(EORALP), licensed under the LGPLv3+.
Copyright (C) 2015 Walter Doekes, OSSO B.V.
"""


class ExactElement(object):
    def __init__(self, api=None, **kwargs):
        super(ExactElement, self).__init__(**kwargs)

        self._api = api

    def get_guid(self):
        # Raises ObjectDoesNotExist if it cannot be found remotely.
        raise NotImplementedError()

    def commit(self):
        # This needs API specific work.
        raise NotImplementedError()



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
Data structures for communication with remote.

This file is part of the Exact Online REST API Library in Python
(EORALP), licensed under the LGPLv3+.
Copyright (C) 2015 Walter Doekes, OSSO B.V.
"""
from .invoice import ExactInvoice       # noqa
from .relation import ExactCustomer     # noqa



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
Standard exceptions.

This file is part of the Exact Online REST API Library in Python
(EORALP), licensed under the LGPLv3+.
Copyright (C) 2015 Walter Doekes, OSSO B.V.
"""


class ExactOnlineError(Exception):
    pass


class MultipleObjectsReturned(ExactOnlineError):
    """
    Exception named Django style.
    """
    pass


class ObjectDoesNotExist(ExactOnlineError):
    """
    Exception named Django style.
    """
    pass



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
Data structures for communication with remote.

Example usage::

    class BossoExactInvoice(ExactInvoice):
        def __init__(self, bosso_invoice=None, **kwargs):
            super(BossoExactInvoice, self).__init__(**kwargs)

            self._bosso_invoice = bosso_invoice

        def get_customer(self):
            return BossoExactCustomer(
                bosso_relation=self._bosso_invoice.relation, api=self._api)

        def get_created_data(self):
            return self._bosso_invoice.open_date

        # ...

This file is part of the Exact Online REST API Library in Python
(EORALP), licensed under the LGPLv3+.
Copyright (C) 2015 Walter Doekes, OSSO B.V.
"""
from .base import ExactElement
from ..exceptions import ExactOnlineError, ObjectDoesNotExist


class UnknownLedgerCodes(ExactOnlineError):
    def __init__(self, ledger_codes):
        super(UnknownLedgerCodes, self).__init__(
            'One or more ledger codes are not found in Exact: %s' % (
                ', '.join(ledger_codes),))
        self.ledger_codes = ledger_codes


class ExactInvoice(ExactElement):
    def get_guid(self):
        exact_invoice = self.__get_remote()
        return exact_invoice['EntryID']

    def get_customer(self):
        # The ExactOnline customer id.
        raise NotImplementedError()

    def get_created_date(self):
        # Period used for both EntryDate and ReportingPeriod. Use the
        # "open_date" if available.
        return NotImplementedError()

    def get_exact_journal(self):
        # E.g. "70" for "Verkoopboek"
        raise NotImplementedError()

    def get_ledger_code_to_guid_map(self, codes):
        """
        Convert set of human codes and to a dict of code to exactonline
        guid mappings.

        Example::

            ret = inv.get_ledger_code_to_guid_map(['1234', '5555'])
            ret == {'1234': '<guid1_from_exactonline_ledgeraccounts>',
                    '5555': '<guid2_from_exactonline_ledgeraccounts>'}
        """
        if codes:
            codes = set(str(i) for i in codes)
            ledger_ids = self._api.ledgeraccounts.filter(code__in=codes)
            ret = dict((str(i['Code']), i['ID']) for i in ledger_ids)
            found = set(ret.keys())
            missing = (codes - found)
            if missing:
                raise UnknownLedgerCodes(missing)
            return ret
        return {}

    def get_ledger_lines(self):
        # Return a bunch of ledger lines, like this:
        # {'code': '1234',          # ledger code
        #  'vat_percentage': '21',  # 21%
        #  'total_amount_excl_vat': Decimal(12.5),
        #  'description': '200 items of foo bar'}
        raise NotImplementedError()

    def get_invoice_number(self):
        raise NotImplementedError()

    def get_total_amount_incl_vat(self):
        # Used in AmountDC (default currency) and AmountFC (foreign
        # currency).
        raise NotImplementedError()

    def get_total_vat(self):
        raise NotImplementedError()

    def hint_exact_invoice_number(self):
        # Exact does not honor all requests.
        raise NotImplementedError()

    def assemble(self):
        invoice_number = self.get_invoice_number()
        customer = self.get_customer()

        total_amount_incl_vat = self.get_total_amount_incl_vat()
        total_vat = self.get_total_vat()
        created_date = self.get_created_date()
        description = u'%s - %s, %s' % (invoice_number, customer.get_name(),
                                        created_date.strftime('%m-%Y'))

        # Make sure the customer exists.
        try:
            customer_guid = customer.get_guid()
        except ObjectDoesNotExist:
            customer.commit()
            customer_guid = customer.get_guid()

        # Compile data to send.
        data = {
            # Converting to string is better than converting to float.
            'AmountDC': str(total_amount_incl_vat),  # DC=default_currency
            'AmountFC': str(total_amount_incl_vat),  # FC=foreign_currency

            # Strange! We receive the date(time) objects as
            # '/Date(unixmilliseconds)/' (mktime(d.timetuple())*1000),
            # but we must send them as ISO8601.
            # NOTE: invoice.open_date is a date, not a datetime, so
            # tzinfo calculations won't work on it, and we cannot use
            # '%z' in the strftime format. Unused code:
            #   import pytz; tzinfo = pytz.timezone('Europe/Amsterdam')
            #   entry_date = tzinfo.localize(invoice.open_date)
            # Pretend we're in UTC and send "Z" zone.
            'EntryDate': created_date.strftime('%Y-%m-%dT%H:%M:%SZ'),

            'Customer': customer_guid,
            'Description': description,
            'Journal': self.get_exact_journal(),
            'ReportingPeriod': created_date.month,
            'ReportingYear': created_date.year,
            'SalesEntryLines': self.assemble_lines(),
            'VATAmountDC': str(total_vat),  # str>float, DC=default_currency
            'VATAmountFC': str(total_vat),  # str>float, FC=foreign_currency
            'YourRef': invoice_number,

            'InvoiceNumber': self.hint_exact_invoice_number(),
        }

        return data

    def assemble_lines(self):
        ret = []

        # Fetch ledger lines.
        ledger_lines = self.get_ledger_lines()

        # Cache ledger codes to ledger GUIDs.
        ledger_ids = self.get_ledger_code_to_guid_map(
            set(i['code'] for i in ledger_lines))

        for ledger_line in self.get_ledger_lines():
            # We must use VATCode. It accepts VATPercentages, but only
            # when it is higher than 0. Not using:
            # 'VATPercentage': str(ledger_line['vat_percentage'] / 100)
            if ledger_line['vat_percentage'] == 0:
                vatcode = '0  '  # FIXME: hardcoded.. fetch from API?
            elif ledger_line['vat_percentage'] == 21:
                vatcode = '2  '  # FIXME: hardcoded.. fetch from API?
            else:
                raise NotImplementedError(
                    'Unknown VAT: %s' % (ledger_line['vat_percentage'],))

            # Again, convert from decimal to str to get more precision.
            line = {'AmountDC': str(ledger_line['total_amount_excl_vat']),
                    'AmountFC': str(ledger_line['total_amount_excl_vat']),
                    'Description': ledger_line['description'],
                    'GLAccount': ledger_ids[ledger_line['code']],
                    'VATCode': vatcode}
            ret.append(line)

        return ret

    def commit(self):
        try:
            exact_guid = self.get_guid()
        except ObjectDoesNotExist:
            exact_guid = None

        data = self.assemble()

        if exact_guid:
            # We cannot supply the SalesEntryLines in the dict like
            # usual when we update (PUT). Instead, we add the new ones
            # and remove the old ones by hand.
            #
            # You don't want to do that the other way around, because
            # deleting the last item removes the entire invoice.
            old_salesentrylines = self.__get_remote()['SalesEntryLines']

            # Add new (and pop the SalesEntryLines from the invoice
            # dict).
            new_salesentrylines = data.pop('SalesEntryLines')
            for line in new_salesentrylines:
                line['EntryID'] = exact_guid
                self._api.restv1('POST', 'salesentry/SalesEntryLines', line)

            # Remove old.
            for line in old_salesentrylines:
                self._api.restv1(
                    'DELETE', ("salesentry/SalesEntryLines(guid'%s')" %
                               line['ID']))

            # Update the invoice (without the SalesEntryLines).
            ret = self._api.invoices.update(exact_guid, data)
            # ret is None
        else:
            ret = self._api.invoices.create(data)
            # ret is a exact_invoice

        # Drop cache, if used.
        if hasattr(self, '_cached_remote'):
            del self._cached_remote

        return ret

    def __get_remote(self):
        if not hasattr(self, '_cached_remote'):
            try:
                self._cached_remote = self._api.invoices.get(
                    invoice_number=self.get_invoice_number())
            except ObjectDoesNotExist as e:
                self._cached_remote = e
        if isinstance(self._cached_remote, Exception):
            raise self._cached_remote
        return self._cached_remote



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
Refreshes OAuth tokens as-needed on receiving a 401.

This file is part of the Exact Online REST API Library in Python
(EORALP), licensed under the LGPLv3+.
Copyright (C) 2015 Walter Doekes, OSSO B.V.
"""
from ..http import HTTPError


class Autorefresh(object):
    def rest(self, method, resource, data=None):
        try:
            decoded = super(Autorefresh, self).rest(method, resource, data=data)
        except HTTPError as e:
            if e.code != 401:
                raise

            # Refresh token.
            self.refresh_token()

            # Retry.
            decoded = super(Autorefresh, self).rest(method, resource, data=data)

        return decoded



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
Autoselects the division.

This file is part of the Exact Online REST API Library in Python
(EORALP), licensed under the LGPLv3+.
Copyright (C) 2015 Walter Doekes, OSSO B.V.
"""
from ..exceptions import ExactOnlineError
from ..http import urljoin
from ..storage import NoOptionError


class V1DivisionError(ExactOnlineError):
    pass


class V1Division(object):
    def restv1(self, method, resource, data=None):
        try:
            division = self.storage.get_division()
        except NoOptionError:
            raise V1DivisionError('Division unset/blank in config')
        if not division:
            raise V1DivisionError('Division unset/blank in config')

        urlbase = 'v1/%d/' % (division,)
        resource = urljoin(urlbase, resource)
        return self.rest(method, resource, data=data)

    def get_divisions(self):
        """
        Get the "current" division and return a dictionary of divisions
        so the user can select the right one.
        """
        ret = self.rest('GET', 'v1/current/Me?$select=CurrentDivision')
        current_division = ret[0]['CurrentDivision']
        assert isinstance(current_division, int)

        urlbase = 'v1/%d/' % (current_division,)
        resource = urljoin(urlbase, 'hrm/Divisions?$select=Code,Description')
        ret = self.rest('GET', resource)

        choices = dict((i['Code'], i['Description']) for i in ret)
        return choices, current_division

    def set_division(self, division):
        """
        Select the "current" division that we'll be working on/with.
        """
        try:
            division = int(division)
        except (TypeError, ValueError):
            raise V1DivisionError('Supplied division %r is not a number' %
                                  (division,))

        urlbase = 'v1/%d/' % (division,)
        resource = urljoin(
            urlbase,
            "crm/Accounts?$select=ID&$filter=Name+eq+'DOES_NOT_EXIST'")
        try:
            self.rest('GET', resource)
        except AssertionError:
            raise V1DivisionError('Invalid division %r according to server' %
                                  (division,))

        self.storage.set_division(division)



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
Helper for receivables (invoices sent by you).

This file is part of the Exact Online REST API Library in Python
(EORALP), licensed under the LGPLv3+.
Copyright (C) 2015 Walter Doekes, OSSO B.V.
"""
from .manager import Manager


class Receivables(Manager):
    """
    Get the elements that make up the Financial Reporting of Outstanding
    Receivables, you call this.
    """
    resource = 'read/financial/ReceivablesList'

    def filter(self, relation_id=None, duedate__lt=None, duedate__gte=None,
               **kwargs):
        """
        A common query would be duedate__lt=date(2015, 1, 1) to get all
        Receivables that are due in 2014 and earlier.
        """
        if relation_id is not None:
            # Filter by (relation) account_id. There doesn't seem to be
            # any reason to prefer
            # 'read/financial/ReceivablesListByAccount?accountId=X' over
            # this.
            relation_id = self._remote_guid(relation_id)
            self._filter_append(kwargs, u'AccountId eq %s' % (relation_id,))

        if duedate__lt is not None:
            # Not sure what the AgeGroup means in
            # ReceivablesListByAgeGroup, but we can certainly do
            # without.
            duedate__lt = self._remote_datetime(duedate__lt)
            self._filter_append(kwargs, u'DueDate lt %s' % (duedate__lt,))

        if duedate__gte is not None:
            # Not sure what the AgeGroup means in
            # ReceivablesListByAgeGroup, but we can certainly do
            # without.
            duedate__gte = self._remote_datetime(duedate__gte)
            self._filter_append(kwargs, u'DueDate ge %s' % (duedate__gte,))

        return super(Receivables, self).filter(**kwargs)



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
HTTP shortcut tests, taken from osso-djuty.

This file is part of the Exact Online REST API Library in Python
(EORALP), licensed under the LGPLv3+.
Copyright (C) 2015-2016 Walter Doekes, OSSO B.V.

We may want to replace this with something simpler.
"""
import ssl
import sys

from os import path
from unittest import TestCase, skipIf

from .http import (
    BadProtocol, HTTPError, Options,
    binquote,
    http_get, http_post, http_put, http_delete,
    opt_secure__unmodified as opt_secure)

try:
    from urllib import request
except ImportError:  # python2
    import urllib2 as request


class HttpTestServer(object):
    "Super simple builtin HTTP test server."
    def __init__(self, method, code, body=None, use_ssl=False):
        from multiprocessing import Process
        from socket import socket

        self.method = method
        self.code = code
        self.body = body

        self.socket = socket()
        if use_ssl:
            here = path.dirname(__file__)
            self.socket = ssl.wrap_socket(
                self.socket,
                certfile=path.join(here, 'http_testserver.crt'),
                keyfile=path.join(here, 'http_testserver.key'))
        self.socket.bind(('127.0.0.1', 0))
        # We must listen() before starting the subprocess, otherwise
        # the main process may get a connection refused if it's too
        # fast.
        self.socket.listen(0)
        self.port = self.socket.getsockname()[1]

        self.process = Process(target=self.respond)
        self.process.start()
        self.socket.close()  # client is done with it

    def join(self):
        self.process.join()

    def respond(self):
        try:
            peersock, peeraddr = self.socket.accept()
        except ssl.SSLError:
            # Broken connection by peer.
            return

        data = peersock.recv(4096)
        if HttpTestCase.to_str(data).startswith(self.method + ' '):
            if self.body is None:
                # If body is None, pass the indata as outdata.
                body = data
                if str != bytes:
                    body = body.decode('utf-8')
            else:
                body = self.body

            peersock.send(
                ('HTTP/1.0 %s Unused Response Title\r\n'
                 'Content-Type: text/plain; utf-8\r\n'
                 '\r\n%s' % (self.code, body)
                 ).encode('utf-8'))
        else:
            peersock.send(
                ('HTTP/1.0 405 Method Not Implemented\r\n'
                 'Content-Type: text/plain; utf-8\r\n'
                 '\r\nUnexpected stuff'
                 ).encode('utf-8'))
        peersock.close()
        self.socket.close()  # server is done with it


class HttpTestCase(TestCase):
    def test_options_or_operator(self):
        a = Options()
        a.protocols = ('ftp',)
        a.cacert_file = 'overwrite_me'
        b = Options()
        b.cacert_file = '/tmp/test.crt'
        c = a | b

        self.assertEqual(c.protocols, ('ftp',))
        self.assertFalse(c.verify_cert)
        self.assertEqual(c.cacert_file, '/tmp/test.crt')

    def test_testserver(self):
        # Ensure that the testserver refuses if the method is bad.
        server = HttpTestServer('FAIL', '555', 'failure')
        self.assertRaises(HTTPError, http_get,
                          'http://127.0.0.1:%d/path' % (server.port,))
        server.join()

    def test_delete(self):
        server = HttpTestServer('DELETE', '200', 'whatever1')
        data = http_delete('http://127.0.0.1:%d/path' % (server.port,))
        server.join()
        self.assertDataEqual(data, 'whatever1')

    def test_get(self):
        server = HttpTestServer('GET', '200', 'whatever2')
        data = http_get('http://127.0.0.1:%d/path' % (server.port,))
        server.join()
        self.assertDataEqual(data, 'whatever2')

    def test_post(self):
        server = HttpTestServer('POST', '200', 'whatever3')
        data = http_post('http://127.0.0.1:%d/path' % (server.port,))
        server.join()
        self.assertDataEqual(data, 'whatever3')

    def test_post_actual_data(self):
        server = HttpTestServer('POST', '200', body=None)  # no body => echo
        indata = 'abc DEF\nghi JKL\n'
        data = http_post(
            'http://127.0.0.1:%d/path' % (server.port,), data=indata)
        server.join()

        data = data.decode('utf-8')
        header, outdata = data.split('\r\n\r\n', 1)
        self.assertEqual(outdata, indata)

    def test_put(self):
        server = HttpTestServer('PUT', '200', 'whatever4')
        data = http_put('http://127.0.0.1:%d/path' % (server.port,))
        server.join()
        self.assertDataEqual(data, 'whatever4')

    def test_502(self):
        server = HttpTestServer('GET', '502', 'eRrOr')
        try:
            http_get('http://127.0.0.1:%d/path' % (server.port,))
        except HTTPError as e:
            self.assertTrue(isinstance(e, request.HTTPError))
            self.assertEqual(e.code, 502)
            self.assertDataEqual(e.response, 'eRrOr')
        else:
            self.assertFalse(True)
        server.join()

    def test_exception_str(self):
        server = HttpTestServer('POST', '503', '{"errno":1}')
        url = 'http://127.0.0.1:%d/path' % (server.port,)
        try:
            http_post(url, data='{"action":1}')
        except HTTPError as e:
            self.assertTrue(isinstance(e, request.HTTPError))
            error_str = str(e)
            self.assertIn('503', error_str)
            self.assertIn(url, error_str)
            self.assertIn('{"action":1}', error_str)
            self.assertIn('{"errno":1}', error_str)
        else:
            self.assertFalse(True)
        server.join()

    def test_https_only_through_options(self):
        self.assertRaises(BadProtocol, http_get,
                          'http://127.0.0.1/path', opt=opt_secure)
        self.assertRaises(BadProtocol, http_get,
                          'ftp://127.0.0.1/path', opt=opt_secure)

    @skipIf(sys.version_info >= (2, 7, 9),
            'PEP-0476: Since Python 2.7.9, certificate verification is always '
            'enabled.')
    def test_https_no_secure(self):
        server = HttpTestServer('GET', '200', 'ssl', use_ssl=True)
        data = http_get('https://127.0.0.1:%d/path' % (server.port,))
        server.join()
        self.assertDataEqual(data, 'ssl')

    def test_https_with_real_secure(self):
        # This should work with a proper certificate.
        data = http_get('https://api.github.com/', opt=opt_secure)
        self.assertEqual(HttpTestCase.to_str(data)[0:1], '{')  # json :)

    def test_https_with_self_signed(self):
        # This should fail, because the testserver uses a self-signed
        # certificate.
        server = HttpTestServer('GET', '200', 'ssl', use_ssl=True)
        self.assertRaises(request.URLError, http_get,
                          'https://127.0.0.1:%d/path' % (server.port,),
                          opt=opt_secure)
        server.join()

    def test_https_with_allowed_self_signed(self):
        my_opt = Options()
        my_opt.cacert_file = path.join(
            path.dirname(__file__), 'http_testserver.crt')
        my_opt = opt_secure | my_opt
        server = HttpTestServer('GET', '200', 'ssl2', use_ssl=True)
        data = http_get('https://127.0.0.1:%d/path' % (server.port,),
                        opt=my_opt)
        server.join()
        self.assertDataEqual(data, 'ssl2')

    def test_https_with_disallowed_real_secure(self):
        # This should fail because we use a custom cacert file which won't
        # contain the real cert.
        my_opt = Options()
        my_opt.cacert_file = path.join(
            path.dirname(__file__), 'http_testserver.crt')
        my_opt = opt_secure | my_opt
        self.assertRaises(request.URLError, http_get,
                          'https://api.github.com/', opt=my_opt)

    # ; Python23 compatibility helpers

    try:
        unicode  # python2
    except NameError:
        to_str = staticmethod(lambda x: x.decode('utf-8'))  # binstr-to-unistr
    else:
        to_str = staticmethod(lambda x: x)  # binstr-to-binstr

    def assertDataEqual(self, data, unistr):
        """
        Compares http_* returned data with a plain string. That plain string
        can be unicode (python3) or binstring (python2). The data is always a
        binstring.
        """
        unistr_passed = self.to_str(data)
        self.assertEqual(unistr_passed, unistr)


class UtilTestCase(TestCase):
    def test_binquote(self):
        self.assertEqual(binquote('abc'), 'abc')
        self.assertEqual(binquote('abc/def'), 'abc/def')  # don't touch /
        self.assertEqual(binquote('abc=def'), 'abc%3Ddef')  # do touch =



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
Base API interface.

This file is part of the Exact Online REST API Library in Python
(EORALP), licensed under the LGPLv3+.
Copyright (C) 2015-2016 Walter Doekes, OSSO B.V.
"""
import json

from time import time

from .http import (
    Options, opt_secure, http_delete, http_get, http_post, http_put,
    binquote, urljoin)


def _json_safe(data):
    """
    json.loads wants an unistr in Python3. Convert it.
    """
    if not hasattr(data, 'encode'):
        try:
            data = data.decode('utf-8')
        except UnicodeDecodeError:
            raise ValueError(
                'Expected valid UTF8 for JSON data, got %r' % (data,))
    return data


class ExactRawApi(object):
    def __init__(self, storage, **kwargs):
        super(ExactRawApi, self).__init__(**kwargs)
        self.storage = storage

    def create_auth_request_url(self):
        # Build the URLs manually so we get consistent order.
        auth_params = {
            'client_id': binquote(self.storage.get_client_id()),
            'redirect_uri': binquote(self.storage.get_response_url()),
            'response_type': 'code',  # or 'token' for JS apps
        }
        auth_data = ('client_id=%(client_id)s'
                     '&redirect_uri=%(redirect_uri)s'
                     '&response_type=%(response_type)s' %
                     auth_params)

        url = '?'.join([self.storage.get_auth_url(), auth_data])
        return url

    def request_token(self, code):
        # Build the URLs manually so we get consistent order.
        token_params = {
            'client_id': binquote(self.storage.get_client_id()),
            'client_secret': binquote(self.storage.get_client_secret()),
            'code': binquote(code),
            'grant_type': 'authorization_code',
            'redirect_uri': binquote(self.storage.get_response_url()),
        }
        token_data = ('client_id=%(client_id)s'
                      '&client_secret=%(client_secret)s'
                      '&code=%(code)s'
                      '&grant_type=%(grant_type)s'
                      '&redirect_uri=%(redirect_uri)s' %
                      token_params)

        # Fire away!
        url = self.storage.get_token_url()
        response = _json_safe(http_post(url, token_data, opt=opt_secure))

        # Validate and store the values.
        self._set_tokens(response)
        # Store the code first after _set_tokens() has validated the
        # data. We don't want to store some bogus code fed to use by Joe
        # Random user.
        self.storage.set_code(code)

    def refresh_token(self):
        # Bring on the fresh stuff!
        # Build the URLs manually so we get consistent order.
        refresh_params = {
            'client_id': binquote(self.storage.get_client_id()),
            'client_secret': binquote(self.storage.get_client_secret()),
            'grant_type': 'refresh_token',
            'refresh_token': binquote(self.storage.get_refresh_token()),
        }
        refresh_data = ('client_id=%(client_id)s'
                        '&client_secret=%(client_secret)s'
                        '&grant_type=%(grant_type)s'
                        '&refresh_token=%(refresh_token)s' %
                        refresh_params)

        # Fire away!
        url = self.storage.get_refresh_url()
        response = _json_safe(http_post(url, refresh_data, opt=opt_secure))

        # Validate and store the values.
        self._set_tokens(response)

    # Don't pass "/api" in the resource, it's in the base URL already!
    # And don't start with a slash either, since we use urljoin on it.
    #
    # See this for a list of possible resources.
    # https://start.exactonline.co.uk/docs/HlpRestAPIResources.aspx?SourceAction=10
    def rest(self, method, resource, data=None):
        url = urljoin(self.storage.get_rest_url().rstrip('/') + '/', resource)

        # Convert data to json.
        if data is None:
            pass
        elif isinstance(data, str):
            pass
        else:
            data = json.dumps(data)

        response = self._rest_query(method, url, data)

        if method in ('DELETE', 'PUT'):
            if response != '':
                raise ValueError('Expected empty data for %s operation: '
                                 'resource=%r, returned=%r' %
                                 (method, resource, response))
            decoded = None
        else:
            try:
                decoded = json.loads(response)
            except ValueError:
                raise ValueError('Expected valid JSON data for %s operation: '
                                 'resource=%r, returned=%r' %
                                 (method, resource, response))

        return decoded

    def _rest_query(self, method, url, data):
        token = self.storage.get_access_token()
        opt_custom = Options()
        opt_custom.headers = {
            'Accept': 'application/json',
            'Authorization': 'Bearer %s' % (token,),
        }
        if method in ('POST', 'PUT'):
            opt_custom.headers.update({'Content-Type': 'application/json'})

        if method == 'DELETE':
            assert data is None
            response = http_delete(url, opt=(opt_secure | opt_custom))
        elif method == 'GET':
            assert data is None
            response = http_get(url, opt=(opt_secure | opt_custom))
        elif method == 'POST':
            response = http_post(url, data, opt=(opt_secure | opt_custom))
        elif method == 'PUT':
            response = http_put(url, data, opt=(opt_secure | opt_custom))
        else:
            raise NotImplementedError('No REST handler for method %s' %
                                      (method,))

        return _json_safe(response)

    def _set_tokens(self, jsondata):
        # The json should look somewhat like this:
        # {"access_token":"AAEA..",
        #  "token_type":"bearer",
        #  "expires_in":"600",
        #  "refresh_token":"__1P!I.."}
        decoded = json.loads(jsondata)

        # Validate the values.
        assert decoded['access_token']
        expires_in = int(decoded['expires_in'])
        assert expires_in > 0
        assert decoded['refresh_token']
        assert decoded['token_type'] == 'bearer'

        # Store the values.
        self.storage.set_access_expiry(int(time()) + expires_in)
        self.storage.set_access_token(decoded['access_token'])
        self.storage.set_refresh_token(decoded['refresh_token'])



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
Helper for ledgeraccount resources.

This file is part of the Exact Online REST API Library in Python
(EORALP), licensed under the LGPLv3+.
Copyright (C) 2015 Walter Doekes, OSSO B.V.
"""
from .manager import Manager


class LedgerAccounts(Manager):
    resource = 'financial/GLAccounts'

    def filter(self, code__in=None, **kwargs):
        # $select=ID,Code,Name
        if 'select' not in kwargs:
            kwargs['select'] = 'ID,Code'

        if code__in is not None:
            code_filter = []
            for code in code__in:
                code = self._remote_code(code)
                code_filter.append(u'Code eq %s' % (code,))
            self._filter_append(kwargs, u'(%s)' % (u' or '.join(code_filter),))
        return super(LedgerAccounts, self).filter(**kwargs)

    def _remote_code(self, code):
        return u"'%s'" % (code.replace("'", "''"),)



#!/usr/bin/env python
from distutils.core import setup

if __name__ == '__main__':
    long_descriptions = []
    with open('README.rst') as file:
        long_descriptions.append(file.read())
    with open('CHANGES.rst') as file:
        long_descriptions.append(file.read())

    setup(
        name='exactonline',
        version='0.2.5',
        packages=['exactonline', 'exactonline.api', 'exactonline.elements'],
        data_files=[('', ['LICENSE.txt', 'README.rst', 'CHANGES.rst'])],
        description='Exact Online REST API Library in Python',
        long_description=('\n\n\n'.join(long_descriptions)),
        author='Walter Doekes, OSSO B.V.',
        author_email='wjdoekes+exactonline@osso.nl',
        url='https://github.com/ossobv/exactonline',
        license='LGPLv3+',
        platforms=['linux'],
        classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Intended Audience :: Developers',
            ('License :: OSI Approved :: GNU Lesser General Public License v3 '
             'or later (LGPLv3+)'),
            'Operating System :: POSIX :: Linux',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Topic :: Office/Business :: Financial',
            'Topic :: Office/Business :: Financial :: Accounting',
            'Topic :: Software Development :: Libraries',
        ],
    )

# vim: set ts=8 sw=4 sts=4 et ai tw=79:



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
Looks for __next values, and begins to download those documents
automatically; it unpaginates resultsets.

This file is part of the Exact Online REST API Library in Python
(EORALP), licensed under the LGPLv3+.
Copyright (C) 2015-2016 Walter Doekes, OSSO B.V.
"""


class Unwrap(object):
    def rest(self, method, resource, data=None):
        iteration = 0
        ret = []

        while resource:
            if iteration >= 50:
                raise ValueError(
                    'Iteration %d limit reached! Last resource %r' % (
                        iteration, resource))

            decoded = super(Unwrap, self).rest(method, resource, data=data)

            # DELETE and PUT methods return None.
            if not decoded:
                assert method in ('DELETE', 'PUT'), method
                assert iteration == 0, iteration
                return decoded

            # GET/POST methods return a host of different types.
            result_data = decoded.pop(u'd', None)
            if result_data is None or decoded:
                raise ValueError(
                    'Expected *only* "d" in response, got this: '
                    'response=%r, d=%r' % (decoded, result_data))

            # POST methods return a nice dictionary inside of 'd'.
            if method == 'POST':
                assert iteration == 0, iteration
                return result_data

            # GET methods...
            assert method == 'GET'

            if isinstance(result_data, dict):
                result_data, resource = self._rest_to_result_data_and_next(
                    result_data)
                ret.extend(result_data)

            elif isinstance(result_data, list):
                assert iteration == 0, iteration
                ret = result_data
                resource = None  # no next

            else:
                raise ValueError(
                    'Expected *list* or *dict* in "d", got this: d=%r' % (
                        result_data,))

            iteration += 1

        return ret

    def _rest_to_result_data_and_next(self, result_data):
        results = result_data.pop(u'results', None)
        next_ = result_data.pop(u'__next', None)
        if results is None or result_data:
            raise ValueError(
                'Expected *only* "results" in "d", got this: '
                'd=%r, results=%r' % (result_data, results))

        ret = results
        assert isinstance(ret, list)
        return ret, next_  # next_ is None when at the end



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
Provides a storage class to the Exact Online REST API Library.

This file is part of the Exact Online REST API Library in Python
(EORALP), licensed under the LGPLv3+.
Copyright (C) 2015 Walter Doekes, OSSO B.V.

Usage:

    storage = IniStorage('somefile.ini')

Or, if you want to use your own storage backend, you may provide a set()
and get() method and inherit from the ExactOnlineConfig.

    class MyStorage(ExactOnlineConfig):
        def get(self, section, option):
            try:
                return self.fetch_stuff_from_somewhere(...)
            except:
                raise NoOptionError()

        def set(self, section, option, value):
            self.store_stuff_somewhere(...)

Example ini file:

    [server]
    auth_url = https://start.exactonline.co.uk/api/oauth2/auth
    rest_url = https://start.exactonline.co.uk/api
    token_url = https://start.exactonline.co.uk/api/oauth2/token

    [application]
    base_url = https://example.com
    client_id = {12345678-abcd-1234-abcd-0123456789ab}
    client_secret = ZZZ999xxx000

    [transient]
    access_expiry = 1426492503
    access_token = dAfjGhB1k2tE2dkG12sd1Ff1A1fj2fH2Y1j1fKJl2f1sD1ON275zJNUy...
    code = dAfj!hB1k2tE2dkG12sd1Ff1A1fj2fH2Y1j1fKJl2f1sD1ON275zJNUy...
    division = 123456
    refresh_token = SDFu!12SAah-un-56su-1fj2fH2Y1j1fKJl2f1sDfKJl2f1sD11FfUn1...

"""
from unittest import TestCase, main
from io import StringIO
from os import path, unlink

from .storage import IniStorage, NoOptionError


class IniStorageTestCase(TestCase):
    def test_dont_die_if_config_doesnt_exist(self):
        try:
            config = IniStorage('storage-does-not-exist.ini')
            del config
            self.assertFalse(path.exists('storage-does-not-exist.ini'))
        finally:
            try:
                unlink('storage-does-not-exist.ini')
            except OSError:
                pass

    def test_missing_section_raises_nooption(self):
        config = IniStorage(StringIO())
        self.assertRaises(NoOptionError, config.get, 'bad_section',
                          'bad_value')

    def test_server_defaults(self):
        config = IniStorage(StringIO())
        self.assertEqual(config.get_auth_url(),
                         'https://start.exactonline.nl/api/oauth2/auth')
        self.assertEqual(config.get_rest_url(),
                         'https://start.exactonline.nl/api')
        self.assertEqual(config.get_token_url(),
                         'https://start.exactonline.nl/api/oauth2/token')

    def test_server_default_writes(self):
        try:
            self.assertFalse(path.exists('storage-gets-created.ini'))
            config = IniStorage('storage-gets-created.ini')
            auth_url = 'https://start.exactonline.nl/api/oauth2/auth'
            self.assertEqual(config.get_auth_url(), auth_url)
            with open('storage-gets-created.ini', 'r') as written:
                for line in written:
                    if auth_url in line:
                        break
                else:
                    self.assertFalse('URL %r not found in file' % (auth_url,))
        finally:
            try:
                unlink('storage-gets-created.ini')
            except OSError:
                pass

    def test_application_no_defaults(self):
        config = IniStorage(StringIO())
        self.assertRaises(NoOptionError, config.get_base_url)
        self.assertRaises(NoOptionError, config.get_client_id)
        self.assertRaises(NoOptionError, config.get_client_secret)

    def test_transient_no_defaults(self):
        config = IniStorage(StringIO())
        self.assertRaises(NoOptionError, config.get_access_expiry)
        self.assertRaises(NoOptionError, config.get_access_token)
        self.assertRaises(NoOptionError, config.get_code)
        self.assertRaises(NoOptionError, config.get_division)
        self.assertRaises(NoOptionError, config.get_refresh_token)

    def test_example_ini(self):
        config = IniStorage(path.join(path.dirname(__file__), 'example.ini'))

        # [server]
        self.assertEqual(config.get_auth_url(),
                         'https://start.exactonline.co.uk/api/oauth2/auth')
        self.assertEqual(config.get_rest_url(),
                         'https://start.exactonline.co.uk/api')
        self.assertEqual(config.get_token_url(),
                         'https://start.exactonline.co.uk/api/oauth2/token')
        self.assertEqual(config.get_token_url(), config.get_refresh_url())

        # [application]
        self.assertEqual(config.get_base_url(), 'https://example.com')
        self.assertEqual(config.get_client_id(),
                         '{12345678-abcd-1234-abcd-0123456789ab}')
        self.assertEqual(config.get_client_secret(), 'ZZZ999xxx000')

        # [transient]
        self.assertEqual(config.get_access_expiry(), 1426492503)
        self.assertEqual(
            config.get_access_token(),
            'dAfjGhB1k2tE2dkG12sd1Ff1A1fj2fH2Y1j1fKJl2f1sD1ON275zJNUy...')
        self.assertEqual(
            config.get_code(),
            'dAfj!hB1k2tE2dkG12sd1Ff1A1fj2fH2Y1j1fKJl2f1sD1ON275zJNUy...')
        self.assertEqual(config.get_division(), 123456)
        self.assertEqual(
            config.get_refresh_token(),
            'SDFu!12SAah-un-56su-1fj2fH2Y1j1fKJl2f1sDfKJl2f1sD11FfUn1...')

    def test_transient_writes(self):
        try:
            self.assertFalse(path.exists('storage-gets-created.ini'))
            config = IniStorage('storage-gets-created.ini')
            self.assertRaises(NoOptionError, config.get_division)
            config.set_division(987654321)
            self.assertEqual(config.get_division(), 987654321)
            config.set_division(11223344)

            # Re-open config, and check availability.
            config2 = IniStorage('storage-gets-created.ini')
            self.assertEqual(config2.get_division(), 11223344)
        finally:
            try:
                unlink('storage-gets-created.ini')
            except OSError:
                pass

    def test_get_and_set_must_be_string_types(self):
        # internals: we want str types inserted into set() and returned
        # from get(). Needed for clean py23 compatibility.
        try:
            self.assertFalse(path.exists('storage-gets-created.ini'))
            config = IniStorage('storage-gets-created.ini')

            # Test even when writing the inverse of a str, we still get a
            # native-str type back. For py23 compatibility.
            byte_string = 'abc'.encode('ascii')
            if hasattr(byte_string, 'encode'):
                non_native_string = byte_string.decode('ascii')
            else:
                non_native_string = byte_string

            # Store temp value.
            config.set_access_token(non_native_string)
            value = config.get_access_token()
            self.assertEqual(value, 'abc')
            self.assertEqual(type(value), str)

            # Reload file and check value some more.
            config = IniStorage('storage-gets-created.ini')
            value = config.get_access_token()
            self.assertEqual(value, 'abc')
            self.assertEqual(type(value), str)
        finally:
            try:
                unlink('storage-gets-created.ini')
            except OSError:
                pass


if __name__ == '__main__':
    main()



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
API tests.

This file is part of the Exact Online REST API Library in Python
(EORALP), licensed under the LGPLv3+.
Copyright (C) 2016 Walter Doekes, OSSO B.V.
"""
import json
from unittest import TestCase

from .api import ExactApi
from .http import opt_secure
from .storage import ExactOnlineConfig, NoOptionError

from .http_test import HttpTestServer


# BEWARE: We disable security here! Don't import this unless you're
# running tests!
opt_secure.protocols = ('http', 'https')


class ApiTestCase(TestCase):
    class MemoryStorage(ExactOnlineConfig):
        def __init__(self, server_port, **kwargs):
            self._data = {
                'server': {
                    'auth_url': 'http://127.0.0.1:%d/auth' % (server_port,),
                    'rest_url': 'http://127.0.0.1:%d/api' % (server_port,),
                    'token_url': 'http://127.0.0.1:%d/token' % (server_port,),
                },
                'application': {
                    'client_id': 'CLIENT_ID',
                    'client_secret': 'CLIENT_SECRET',
                },
                'transient': {
                    'division': '1',
                    'access_token': 'ACCESS_TOKEN',
                    'refresh_token': 'REFRESH_TOKEN',
                },
            }
            for key, value in kwargs.items():
                self._data[key].update(**value)

        def get_division(self):
            return 1

        def get(self, section, option):
            if section not in self._data:
                raise NoOptionError(section, option)
            try:
                value = self._data[section][option]
            except KeyError:
                raise NoOptionError(section, option)
            return value

        def set(self, section, option, value):
            if section not in self._data:
                self._data[section] = {}
            self._data[section][option] = value

    def get_api(self, server_port):
        storage = self.MemoryStorage(server_port=server_port)
        return ExactApi(storage=storage)

    def test_call(self):
        data = {
            'd': {'results': [
                {'name': 'invoice1', 'identifier': 4},
                {'name': 'invoice2', 'identifier': 44},
            ]},
        }
        jsondata = json.dumps(data)
        server = HttpTestServer('GET', '200', jsondata)
        api = self.get_api(server_port=server.port)
        res = api.invoices.filter(filter=u"Currency eq '\u20ac'", top=5)
        server.join()
        self.assertEqual(res, data['d']['results'])



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
HTTP shortcuts, taken from osso-djuty.

This file is part of the Exact Online REST API Library in Python
(EORALP), licensed under the LGPLv3+.
Copyright (C) 2015-2016 Walter Doekes, OSSO B.V.

We may want to replace this with something simpler.
"""
import urllib
import socket
import ssl
import sys

try:
    from ssl import create_default_context
except ImportError:  # python2.7.9-
    create_default_context = None
try:
    from http.client import HTTPConnection, HTTPS_PORT
except ImportError:  # python2
    from httplib import HTTPConnection, HTTPS_PORT
try:
    from urllib import request
except ImportError:  # python2
    import urllib2 as request
try:
    from urllib.parse import urljoin, quote
except ImportError:  # python2
    from urllib import quote
    from urlparse import urljoin

# For older Python, use this. For newer Python, use nothing to get
# libssl-selected files instead. You can choose to override this
# by importing exactonline.http and updating this value before calling
# any of the http_* functions (through an exactonline API call).
FALLBACK_CACERT_FILE = '/etc/ssl/certs/ca-certificates.crt'


# ; helpers

def binquote(value):
    """
    We use quote() instead of urlencode() because the Exact Online API
    does not like it when we encode slashes -- in the redirect_uri -- as
    well (which the latter does).
    """
    return quote(value.encode('utf-8'))

urljoin  # touch it, we don't use it


# ; http stuff

class BadProtocol(ValueError):
    """
    Raised when you try to http_get or http_post with a disallowed
    protocol.
    """
    pass


class HTTPError(request.HTTPError):
    """
    Override the original HTTPError, drop the fp and add a response and
    add request method and data.
    """
    def __init__(self, url, code, msg, hdrs, response, reqmethod, reqdata):
        request.HTTPError.__init__(self, url, code, msg, hdrs, None)
        self.url = url
        self.response = response        # in
        self.reqmethod = reqmethod
        self.reqdata = reqdata or ''    # out

    @staticmethod
    def clean_and_trim(value):
        if not isinstance(value, str):
            value = value.decode('utf-8', 'replace')
        value = value[0:1023] + ('', '...')[len(value) > 1023]
        value = ''.join(
            ('?', i)[0x20 <= ord(i) <= 0x7F or i in '\t\n\r'] for i in value)
        return value

    def get_content_type(self):
        try:
            # self.hdrs is a ???
            type_ = str(self.hdrs.type)
        except AttributeError:
            # self.hdrs is a HTTPMessage
            type_ = '|'.join(
                i[1] for i in self.hdrs.items()
                if i[0].lower() == 'content-type')
        return type_

    def format_req(self):
        return '%s %s\nContent-Length: %s\n\n%s' % (
            self.reqmethod, self.url, len(self.reqdata),
            self.clean_and_trim(self.reqdata))

    def format_resp(self):
        return '<RESP> %s %s\nContent-Type: %s\nContent-Length: %s\n\n%s' % (
            self.code, self.msg, self.get_content_type(), len(self.response),
            self.clean_and_trim(self.response))

    def __str__(self):
        return (
            'HTTPError: %(msg)s%(sep)s\n%(out)s%(sep)s\n%(in)s%(sep)s' % {
                'msg': self.msg, 'out': self.format_req(),
                'in': self.format_resp(), 'sep': '\n' + ('-' * 71)})


class Options(object):
    # Which protocols to we allow.
    protocols = ('http', 'https')
    # Do we validate the SSL certificate.
    verify_cert = False
    # What we use to validate the SSL certificate.
    cacert_file = None  # None means "use default or fallback if no default"
    # Optional headers.
    headers = None

    # Which properties we have.
    _PROPERTIES = ('protocols', 'verify_cert', 'cacert_file', 'headers')

    def __or__(self, other):
        """
        Join multiple Options together with the or-operator '|'.
        It uses the identity operator to compare values against the
        default values, so non-overridden values won't overwrite
        overridden ones.

        BUG: This will fail if you try to re-set booleans from False
        to True to False.
        """
        new_options = Options()

        for source in (self, other):
            for item in self._PROPERTIES:
                source_item = getattr(source, item)
                if source_item is not getattr(Options, item):  # identity check
                    setattr(new_options, item, source_item)

        return new_options

# Default options.
opt_default = Options()
opt_default__unmodified = opt_default | Options()  # copy, for testing

# SSL-safe options.
opt_secure = Options()
opt_secure.protocols = ('https',)
opt_secure.verify_cert = True
opt_secure__unmodified = opt_secure | Options()  # copy, for testing


class Request(request.Request):
    """
    Override the request.Request class to supply a custom method.
    """
    def __init__(self, method=None, *args, **kwargs):
        request.Request.__init__(self, *args, **kwargs)
        assert method in ('DELETE', 'GET', 'POST', 'PUT')
        self._method = method

    def get_method(self):
        return self._method


class ValidHTTPSConnection(HTTPConnection):
    """
    This class allows communication via SSL.

    Originally by: Walter Cacau, 2013-01-14
    Source: http://stackoverflow.com/questions/6648952/\
            urllib-and-validation-of-server-certificate
    """
    default_port = HTTPS_PORT
    cacert_file = opt_default.cacert_file

    def __init__(self, *args, **kwargs):
        HTTPConnection.__init__(self, *args, **kwargs)

    def connect(self):
        "Connect to a host on a given (SSL) port."
        sock = socket.create_connection((self.host, self.port),
                                        self.timeout, self.source_address)
        if self._tunnel_host:
            self.sock = sock
            self._tunnel()

        # Python 2.7.9+
        if create_default_context:
            # Newer python will use the "right" cacert file automatically. So
            # the default of None can safely be passed along.
            ctx = create_default_context(cafile=self.cacert_file)
            sock = ctx.wrap_socket(sock, server_hostname=self.host)
        else:
            # Take the supplied file, or FALLBACK_CACERT_FILE if nothing
            # was supplied.
            cacert_file = self.cacert_file or FALLBACK_CACERT_FILE
            sock = ssl.wrap_socket(sock, ca_certs=cacert_file,
                                   cert_reqs=ssl.CERT_REQUIRED)

        self.sock = sock


class ValidHTTPSHandler(request.HTTPSHandler):
    """
    Originally by: Walter Cacau, 2013-01-14
    Source: http://stackoverflow.com/questions/6648952/\
            urllib-and-validation-of-server-certificate
    """
    def __init__(self, cacert_file):
        self.cacert_file = cacert_file
        request.HTTPSHandler.__init__(self)

    def https_open(self, req):
        # If someone uses an alternate cacert_file, we have no decent
        # way of telling that to a subclass (not instance).
        if self.cacert_file == ValidHTTPSConnection.cacert_file:
            class_ = ValidHTTPSConnection
        else:
            # Yuck. Create a local subclass so we can set a custom
            # cacert_file.
            class CustomValidHTTPSConnection(ValidHTTPSConnection):
                cacert_file = self.cacert_file
            class_ = CustomValidHTTPSConnection

        return self.do_open(class_, req)


def http_delete(url, opt=opt_default):
    """
    Shortcut for urlopen (DELETE) + read. We'll probably want to add a
    nice timeout here later too.
    """
    return _http_request(url, method='DELETE', opt=opt)


def http_get(url, opt=opt_default):
    """
    Shortcut for urlopen (GET) + read. We'll probably want to add a nice
    timeout here later too.
    """
    return _http_request(url, method='GET', opt=opt)


def http_post(url, data=None, opt=opt_default):
    """
    Shortcut for urlopen (POST) + read. We'll probably want to add a
    nice timeout here later too.
    """
    return _http_request(url, method='POST', data=_marshalled(data), opt=opt)


def http_put(url, data=None, opt=opt_default):
    """
    Shortcut for urlopen (PUT) + read. We'll probably want to add a nice
    timeout here later too.
    """
    return _http_request(url, method='PUT', data=_marshalled(data), opt=opt)


def _marshalled(data):
    if not data:
        data = ''.encode('utf-8')  # ensure PUT/POST-mode
    elif isinstance(data, type(u'')):
        data = data.encode('utf-8')
    elif isinstance(data, (bytes, str)):
        pass
    else:
        data = urllib.urlencode(data)
    return data


def _http_request(url, method=None, data=None, opt=None):
    # Check protocol.
    proto = url.split(':', 1)[0]
    if proto not in opt.protocols:
        raise BadProtocol('Protocol %s in URL %r disallowed by caller' %
                          (proto, url))

    # Create URL opener.
    if opt.verify_cert:
        # It's legal to pass either a class or an instance here.
        opener = request.build_opener(ValidHTTPSHandler(opt.cacert_file))
    else:
        opener = request.build_opener()

    # Create the Request with optional extra headers.
    req = Request(url=url, data=data, method=method,
                  headers=(opt.headers or {}))

    exc_info, fp, stored_exception = None, None, None
    try:
        fp = opener.open(req)
        # print fp.info()  # (temp, print headers)
        response = fp.read()
    except request.HTTPError as exception:
        fp = exception.fp  # see finally clause
        exc_info = sys.exc_info()
        stored_exception = exception
    except Exception as exception:
        exc_info = sys.exc_info()
        stored_exception = exception
    finally:
        if fp:
            # Try a bit harder to flush the connection and close it
            # properly. In case of errors, our django testserver peer
            # will show an error about us killing the connection
            # prematurely instead of showing the URL that causes the
            # error. Flushing the data here helps.
            if exc_info:
                response = fp.read()
                fp.close()
                # And, even more importantly. Some people want the
                # exception/error info. Store it in our HTTPError
                # subclass.
                raise HTTPError(
                    exc_info[1].url,
                    exc_info[1].code,
                    exc_info[1].msg,
                    exc_info[1].hdrs,
                    response,
                    method or '(none)',
                    data)
            fp.close()

    if exc_info:
        raise stored_exception  # exc_info[0], exc_info[1], exc_info[2]
    return response



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
The Exact Online REST API Library in Python (EORALP), licensed under the
LGPLv3+.
Copyright (C) 2015 Walter Doekes, OSSO B.V.

See README.rst for more info.
"""
